# TODO: write checks for all inputs

from __future__ import absolute_import
from __future__ import print_function

import logging
import numpy as np
from math import factorial
from random import sample, randint
from itertools import permutations
# from networkx import astar_path, NetworkXException
from cspy.preprocessing import check
from cspy.path import Path

log = logging.getLogger(__name__)


class GRASP:
    """
    Greedy Randomised Adaptive Search Procedure for the (resource) constrained
    shortest path problem. Adapted from `Ferone et al 2019`_.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute. Also, the number of nodes must be
        :math:`\geq 5`.

    max_res : list of floats
        :math:`[M_1, M_2, ..., M_{n\_res}]` upper bounds for resource
        usage (including initial forward stopping point).

    min_res : list of floats
        :math:`[L_1, L_2, ..., L_{n\_res}]` lower bounds for resource
        usage (including initial backward stopping point).
        We must have ``len(min_res)`` :math:`=` ``len(max_res)``.

    max_iter : int
        Maximum number of iterations for algorithm

    max_no_improvement : int
        Maximum number of iterations without improvement

    alpha : float, optional
        Greediness factor 0 (random) --> 1 (greedy)

    Returns
    -------
    path : list
        nodes in shortest path obtained.

    Raises
    ------
    Exception
        if no resource feasible path is found

    Notes
    -----
    The input graph must have a ``n_res`` attribute.
    The edges in the graph must all have a ``res_cost`` attribute.
    See `Using cspy`_

    .. _Using cspy: https://cspy.readthedocs.io/en/latest/how_to.html


    Example
    -------
    To run the algorithm, create a :class:`GRASP` instance and call `run`.

    .. code-block:: python

        >>> from cspy import GRASP
        >>> G = DiGraph(directed=True, n_res=2)
        >>> G.add_edge('Source', 'A', res_cost=[1, 1], weight=1)
        >>> G.add_edge('Source', 'B', res_cost=[1, 1], weight=1)
        >>> G.add_edge('Source', 'C', res_cost=[10, 1], weight=10)
        >>> G.add_edge('A', 'C', res_cost=[1, 1], weight=1)
        >>> G.add_edge('A', 'E', res_cost=[10, 1], weight=10)
        >>> G.add_edge('A', 'F', res_cost=[10, 1], weight=10)
        >>> G.add_edge('B', 'C', res_cost=[2, 1], weight=-1)
        >>> G.add_edge('B', 'F', res_cost=[10, 1], weight=10)
        >>> G.add_edge('B', 'E', res_cost=[10, 1], weight=10)
        >>> G.add_edge('C', 'D', res_cost=[1, 1], weight=-1)
        >>> G.add_edge('D', 'E', res_cost=[1, 1], weight=1)
        >>> G.add_edge('D', 'F', res_cost=[1, 1], weight=1)
        >>> G.add_edge('D', 'Sink', res_cost=[10, 10], weight=10)
        >>> G.add_edge('F', 'Sink', res_cost=[10, 1], weight=1)
        >>> G.add_edge('E', 'Sink', res_cost=[1, 1], weight=1)
        >>> path = GRASP(G, [5, 5], [0, 0], 50, 10).run()
        >>> print(path)
        ['Source', 'A', 'C', 'D', 'E', 'Sink']

    .. _Ferone et al 2019: https://www.tandfonline.com/doi/full/10.1080/10556788.2018.1548015

    """

    def __init__(self, G, max_res, min_res,
                 max_iter, max_no_improvement, alpha=0.9):
        # Check input graph and parameters
        check(G, max_res, min_res)
        # Input parameters
        self.G = G
        self.max_res = max_res
        self.min_res = min_res
        self.max_iter = max_iter
        self.max_no_improvement = max_no_improvement
        self.alpha = alpha
        # Algorithm specific parameters
        self.it = 0
        self.best = None
        self.stop = False
        self.nodes = self.G.nodes()
        # list(node for node in self.G.nodes()
        #                   if node != 'Source' and node != 'Sink')

    def run(self):
        while self.it < self.max_iter and not self.stop:
            self.algorithm()
            self.it += 1
        return self.best.list

    def algorithm(self):
        solution = self._construct()
        solution = self._local_search(solution)
        self._update_best(solution)

    def _construct(self):
        # TODO maybe make solution into an object
        solution = Solution(sample(self.nodes, 1), 0)
        while len(solution.list) < len(self.nodes):
            candidates = [i for i in self.nodes
                          if i not in solution.list]
            weights = [self._heuristic(solution.list[-1], i)
                       for i in candidates]
            # Build Restricted Candidiate List
            restriced_candidates = [
                candidates[i] for i, c in enumerate(weights)
                if c <= (min(weights) +
                         self.alpha * (max(weights) - min(weights)))]
            solution.list.append(np.random.choice(restriced_candidates))
            solution.cost = self._cost_solution(solution)
        return solution

    def _local_search(self, solution):
        it = 0  # local iteration counter
        while it < self.max_no_improvement:
            candidate = Solution(
                self._find_alternative_paths(solution.list), 0)
            candidate.cost = self._cost_solution(candidate)
            if (candidate.cost < solution.cost and
                    self._check_path(candidate)):
                solution = candidate
                self.stop = True
            it += 1
        return solution

    def _update_best(self, solution):
        if not self.best or solution.cost < self.best.cost:
            self.best = solution

    def _heuristic(self, i, j):
        # Given a node pair returns a weight to apply
        if i and j:
            if (i, j) not in self.G.edges():
                return 1e10
            else:
                return self.G.get_edge_data(i, j)['weight']
        else:
            return 1e10

    def _cost_solution(self, solution=None):
        if solution:
            return sum(self._heuristic(i, j)
                       for i, j in zip(solution.list,
                                       solution.list[1:]))
        else:
            return 1e11

    def _check_path(self, solution=None):
        if solution:
            path, cost = solution.list, solution.cost
            if (len(path) > 2 and cost < 1e10 and
                    path[0] == 'Source' and
                    path[-1] == 'Sink'):
                edge_or_true = Path(self.G, path, self.max_res,
                                    self.min_res)._check_feasibility()
                if edge_or_true is True:
                    return True
                else:
                    return False
            else:
                return False
        else:
            return False

    def _find_alternative_paths(self, path):
        n_permutations = int(factorial(len(path)) / factorial(len(path) - 2))
        sample_size = randint(1, n_permutations)
        selection = sample(list(permutations(path, 2)), sample_size)
        path_edges = dict(list(edge for edge in selection
                               if edge in self.G.edges()))
        elem = 'Source'  # start point in the new list
        new_list = []
        for _ in range(len(path_edges)):
            try:
                new_list.append((elem, path_edges[elem]))
                elem = path_edges[elem]
            except KeyError:
                pass
        if new_list:
            nodes_to_keep = [t[0] for t in new_list]
            nodes_to_keep.append(new_list[-1][1])
        else:
            nodes_to_keep = []
        return nodes_to_keep


class Solution(object):
    """
    Object for Solutions and candidates during GRASP iterations.

    Parameters
    ----------

    nodes : list
        list of nodes in solution

    cost : float
        cost of solution
    """

    def __init__(self, nodes, cost):
        self.list = nodes
        self.cost = cost

    def __repr__(self):
        return str(self)

    def __str__(self):  # for printing purposes
        return "Solution({0},{1})".format(self.list, self.cost)



import logging
import networkx as nx
from cspy.checks import *

log = logging.getLogger(__name__)


def check(G, max_res=None, min_res=None, direction=None, algorithm=None):
    """Check whether inputs have the appropriate attributes and
    are of the appropriate types."""

    errors = []
    # Select checks to perform based on the input provided
    if max_res and min_res and direction:
        check_funcs = [_check_res, _check_direction, _check_graph_attr,
                       _check_edge_attr, _check_path]
    elif max_res and min_res:
        check_funcs = [_check_res, _check_graph_attr,
                       _check_edge_attr, _check_path]
    else:
        check_funcs = [_check_path]
    # Check all functions in check_funcs
    for func in check_funcs:
        try:
            func(G, max_res, min_res, direction, algorithm)
        except Exception as e:
            errors.append(e)  # if check fails save error message
    if errors:
        # if any check has failed raise an exception with all the errors
        raise Exception('\n'.join('{}'.format(item) for item in errors))


def prune_graph(G, max_res, min_res):
    """Removes nodes that cannot be reached due to resource limits."""

    def _check_resource(r):
        # check resource r's feasibility along a path

        def __get_weight(i, j, attr_dict):
            # returns number to use as weight for the algorithm
            return attr_dict['res_cost'][r]

        # Get paths from source to all other nodes
        length, path = nx.single_source_bellman_ford(
            G, 'Source', weight=__get_weight)
        try:
            # If any path violates the resource upper or lower bounds
            # then, add the problematic node to the dictionary
            nodes_to_remove.update({
                path[key][-2]: val for key, val in length.items()
                if val > max_res[r] or val < min_res[r]})
            if "Source" in nodes_to_remove or "Sink" in nodes_to_remove:
                raise Exception("Sink not reachable for resource {}".format(r))
            # path is a dict of the form:
            # {node_i: [Source, ..., node_k, node_i]}.
            # Hence, if node_i is found to violate a resource constraint, it is
            # because of node_k, therefore, we add path[key][-2] = node_k to
            # the dictionary of nodes to remove.
        except IndexError:  # No nodes violate resource limits
            pass

    nodes_to_remove = {}
    n_nodes = len(G.nodes())
    # Map function for each resource
    list(map(_check_resource, range(0, G.graph['n_res'])))
    if nodes_to_remove:  # if there are nodes to remove
        # Filter out source or sink
        nodes_to_remove = {
            key: val for key, val in nodes_to_remove.items()
            if key != 'Source' or key != 'Sink'}
        G.remove_nodes_from(nodes_to_remove)
        log.info("Removed {0}/{1} nodes".format(len(nodes_to_remove),
                                                n_nodes))
    return G


def check_and_preprocess(preprocess, G, max_res=None, min_res=None,
                         direction=None, algorithm=None):
    """
    Checks whether inputs and the graph are of the appropriate types and
    have the required properties.
    Removes nodes that cannot be reached due to resource limits.

    Parameters
    ----------
    preprocess : bool
        enables preprocessing routine.

    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats, optional
        :math:`[L, M_1, M_2, ..., M_{n\_res}]`
        upper bound for resource usage.
        We must have ``len(max_res)`` :math:`\geq 2`

    min_res : list of floats, optional
        :math:`[U, L_1, L_2, ..., L_{nres}]` lower bounds for resource usage.
        We must have ``len(min_res)`` :math:`=` ``len(max_res)`` :math:`\geq 2`

    direction : string, optional
        preferred search direction. Either 'both','forward', or, 'backward'.
        Default : 'both'.

    :return: If ``preprocess``, returns preprocessed graph ``G`` if no
        exceptions are raised, otherwise doesn't return anything.

    :raises: Raises exceptions if incorrect input is given.
        If multiple exceptions are raised, and exception with a list of
        exceptions is raised.
    """
    check(G, max_res, min_res, direction, algorithm)
    if preprocess:
        G = prune_graph(G, max_res, min_res)
        check(G)
    return G



from numpy import zeros, array


class Path(object):
    """docstring for Path"""

    def __init__(self, G, path, max_res, min_res):
        self.G = G
        self.path = path
        self.max_res = max_res
        self.min_res = min_res

    def __repr__(self):
        return str(self)

    def __str__(self):  # for printing purposes
        return "Path({0},{1},{2})".format(
            self.path, self.max_res, self.min_res)

    def _check_feasibility(self):
        shortest_path_edges = (
            edge for edge in self.G.edges(
                self.G.nbunch_iter(self.path), data=True)
            if edge[0:2] in zip(self.path, self.path[1:])
        )
        total_res = zeros(self.G.graph['n_res'])
        # Check path for resource feasibility by adding one edge at a time
        for edge in shortest_path_edges:
            total_res += self._edge_extract(edge)
            if (all(total_res <= self.max_res) and
                    all(total_res >= self.min_res)):
                pass
            else:
                break
        else:
            return True
        return edge

    @staticmethod
    def _edge_extract(edge):
        return array(edge[2]['res_cost'])



from __future__ import absolute_import
from __future__ import print_function

import logging
import numpy as np
from networkx import astar_path, NetworkXException
from cspy.path import Path
from cspy.label import Label
from cspy.preprocessing import check


class Tabu:
    """
    Simple Tabu-esque algorithm for the (resource) constrained shortest
    path problem.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats
        :math:`[M_1, M_2, ..., M_{n\_res}]` upper bounds for resource
        usage (including initial forward stopping point).

    min_res : list of floats
        :math:`[L_1, L_2, ..., L_{n\_res}]` lower bounds for resource
        usage (including initial backward stopping point).
        We must have ``len(min_res)`` :math:`=` ``len(max_res)``.

    Returns
    -------
    path : list
        nodes in shortest path obtained.

    Notes
    -----
    The input graph must have a ``n_res`` attribute.
    The edges in the graph must all have a ``res_cost`` attribute.
    See `Using cspy`_

    .. _Using cspy: https://cspy.readthedocs.io/en/latest/how_to.html

    Example
    -------
    To run the algorithm, create a :class:`Tabu` instance and call `run`.

    .. code-block:: python

        >>> from cspy import Tabu
        >>> G = nx.DiGraph(directed=True, n_res=2)
        >>> G.add_edge('Source', 'A', res_cost=[1, 1], weight=1)
        >>> G.add_edge('Source', 'B', res_cost=[1, 1], weight=1)
        >>> G.add_edge('A', 'C', res_cost=[1, 1], weight=1)
        >>> G.add_edge('B', 'C', res_cost=[2, 1], weight=-1)
        >>> G.add_edge('C', 'D', res_cost=[1, 1], weight=-1)
        >>> G.add_edge('D', 'E', res_cost=[1, 1], weight=1)
        >>> G.add_edge('D', 'F', res_cost=[1, 1], weight=1)
        >>> G.add_edge('F', 'Sink', res_cost=[1, 1], weight=1)
        >>> G.add_edge('E', 'Sink', res_cost=[1, 1], weight=1)
        >>> max_res, min_res = [5, 5], [0, 0]
        >>> path = Tabu(G, max_res, min_res).run()
        >>> print(path)
        ['Source', 'A', 'C', 'D', 'E', 'Sink']

    """

    def __init__(self, G, max_res, min_res):
        # Check input graph and parameters
        check(G, max_res, min_res)
        # Input parameters
        self.G = G
        self.max_res = max_res
        self.min_res = min_res
        # Algorithm specific parameters
        self.it = 0
        self.path = []
        self.stop = False
        self.neighbour = 'Source'
        self.neighbourhood = []
        self.tabu_edge = None
        self.edges_to_check = dict(self.G.edges())

    def run(self):
        """
        Runs Tabu Search with resource constraints for the input graph
        """
        while self.stop is False:
            self.algorithm()
            self.it += 1

        if self.path:
            logging.debug(self.path)
            return self.path
        else:
            raise Exception("No resource feasible path has been found")

    def algorithm(self):
        path = []
        try:
            path = astar_path(self.G, self.neighbour,
                              'Sink', heuristic=self._heuristic)
        except NetworkXException:
            pass
        if path:
            self._update_path(path)
            edge_or_true = Path(self.G, self.path, self.max_res,
                                self.min_res)._check_feasibility()
            if edge_or_true is True:
                self.stop = True
            else:
                self._get_neighbour(edge_or_true)

            # shortest_path_edges = (
            #     edge for edge in self.G.edges(
            #         self.G.nbunch_iter(self.path), data=True)
            #     if edge[0:2] in zip(self.path, self.path[1:]))
            # total_res = np.zeros(self.G.graph['n_res'])
            # # Check path for resource feasibility by adding one edge at a time
            # for edge in shortest_path_edges:
            #     total_res += self._edge_extract(edge)
            #     if (all(total_res <= self.max_res) and
            #             all(total_res >= self.min_res)):
            #         pass
            #     else:
            #         self._get_neighbour(edge)
            #         break
            # else:
            #     self.stop = True
        else:
            self._get_neighbour(self.tabu_edge)

    def _update_path(self, path):
        # Joins path using previous path and [neighbour, ..., sink] path
        if self.it == 0:
            self.path = path
        if self.neighbour in self.path:
            self.path = [node for node in self.path
                         if (node != self.neighbour and
                             self.path.index(node) <
                             self.path.index(self.neighbour))] + path
        else:
            self._merge_paths(path)

    def _merge_paths(self, path):
        branch_path = [n for n in self.path if n not in path]
        for node in reversed(branch_path):
            if (node, self.neighbour) in self.G.edges():
                self.path = [n for n in branch_path
                             if (branch_path.index(n) <=
                                 branch_path.index(node))] + path
                break

    def _get_neighbour(self, edge=None):
        if self.edges_to_check:
            if edge and edge[:2] in self.edges_to_check:
                # If edge not already been seen
                current_edge = edge
            elif edge[0] == "Source":
                self.stop = True
                return
            else:
                current_edge = self._get_next_neighbour_edge(self.tabu_edge)
        else:
            self.stop = True
            return
        self.edges_to_check.pop(current_edge[:2], None)
        self.neighbour = current_edge[0]
        self.tabu_edge = current_edge

    def _get_next_neighbour_edge(self, edge):
        # Retrieves the edge adjacent to node with the least weight
        if not self.neighbourhood:
            node = edge[0]
            if node == "Source":
                self.stop = True
                return edge
            self.neighbourhood = [
                e for e in self.G.edges(
                    self.G.nbunch_iter(
                        [node] + list(self.G.predecessors(node))),
                    data=True)
                if e[1] == node and e != edge]
            self.neighbourhood.sort(
                key=lambda x: x[2]['weight'])
        next_edge = self.neighbourhood[-1]
        self.neighbourhood.pop(-1)
        return next_edge

    def _heuristic(self, i, j):
        # Given a node pair returns a weight to apply
        if (i, j) == self.tabu_edge or not (i, j) in self.G.edges():
            return 1e10
        else:
            return self.G.get_edge_data(i, j)['weight']

    @staticmethod
    def _edge_extract(edge):
        return np.array(edge[2]['res_cost'])



from __future__ import absolute_import
from __future__ import print_function

import logging
import numpy as np
from networkx import astar_path, NetworkXException
from cspy.path import Path
from cspy.preprocessing import check

log = logging.getLogger(__name__)


class GreedyElim:
    """
    Simple Greedy elimination algorithm for the (resource) constrained shortest
    path problem. The algorithms solves a standard shortest path problem and
    eliminates resource infeasible edges iteratively until a resource feasible
    path is found.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats
        :math:`[M_1, M_2, ..., M_{n\_res}]` upper bounds for resource
        usage (including initial forward stopping point).

    min_res : list of floats
        :math:`[L_1, L_2, ..., L_{n\_res}]` lower bounds for resource
        usage (including initial backward stopping point).
        We must have ``len(min_res)`` :math:`=` ``len(max_res)``.

    return_G : bool, optional
        whether or not you'd like the resulting graph returned

    Returns
    -------
    path : list
        nodes in shortest path obtained.

    Raises
    ------
    Exception
        if no resource feasible path is found

    Notes
    -----
    The input graph must have a ``n_res`` attribute.
    The edges in the graph must all have a ``res_cost`` attribute.
    See `Using cspy`_

    .. _Using cspy: https://cspy.readthedocs.io/en/latest/how_to.html


    Example
    -------
    To run the algorithm, create a :class:`GreedyElim` instance and call `run`.

    .. code-block:: python

        >>> from cspy import GreedyElim
        >>> G = nx.DiGraph(directed=True, n_res=2)
        >>> G.add_edge('Source', 'A', res_cost=[1, 1], weight=1)
        >>> G.add_edge('Source', 'B', res_cost=[1, 1], weight=1)
        >>> G.add_edge('A', 'C', res_cost=[1, 1], weight=1)
        >>> G.add_edge('B', 'C', res_cost=[2, 1], weight=-1)
        >>> G.add_edge('C', 'D', res_cost=[1, 1], weight=-1)
        >>> G.add_edge('D', 'E', res_cost=[1, 1], weight=1)
        >>> G.add_edge('D', 'F', res_cost=[1, 1], weight=1)
        >>> G.add_edge('F', 'Sink', res_cost=[1, 1], weight=1)
        >>> G.add_edge('E', 'Sink', res_cost=[1, 1], weight=1)
        >>> max_res, min_res = [5, 5], [0, 0]
        >>> path = GreedyElim(G, max_res, min_res).run()
        >>> print(path)
        ['Source', 'A', 'C', 'D', 'E', 'Sink']

    """

    def __init__(self, G, max_res, min_res, return_G=False):
        # Check input graph and parameters
        check(G, max_res, min_res)
        # Input parameters
        self.G = G
        self.max_res = max_res
        self.min_res = min_res
        self.return_G = return_G
        # Algorithm specific parameters
        self.it = 0
        self.path = []
        self.stop = False
        self.predecessor_edges = []
        self.last_edge_removed = None
        self.edges_to_remove = dict(self.G.edges())

    def run(self):
        while self.stop is False:
            self.algorithm()
            self.it += 1

        if self.path:
            if self.return_G:
                return self.G, self.path
            else:
                return self.path
        else:
            raise Exception("No resource feasible path has been found")

    def algorithm(self):
        path = []
        try:
            path = astar_path(self.G, 'Source', 'Sink')
        except NetworkXException:
            pass
        if path:
            edge_or_true = Path(self.G, path, self.max_res,
                                self.min_res)._check_feasibility()
            if edge_or_true is True:
                self.path = path
                self.stop = True
            else:
                self._update_graph(edge_or_true)
            # shortest_path_edges = (
            #     edge for edge in self.G.edges(
            #         self.G.nbunch_iter(path), data=True)
            #     if edge[0:2] in zip(path, path[1:]))
            # total_res = np.zeros(self.G.graph['n_res'])
            # # Check path for resource feasibility by adding one edge at a time
            # for edge in shortest_path_edges:
            #     total_res += self._edge_extract(edge)
            #     if (all(total_res <= self.max_res) and
            #             all(total_res >= self.min_res)):
            #         pass
            #     else:
            #         self._update_graph(edge)
            #         break
            # else:  # no break so resource feasible path found
            #     self.path = path
            #     self.stop = True
        else:
            self.G.add_edge(*self.last_edge_removed[:2],
                            res_cost=self.last_edge_removed[2]['res_cost'],
                            weight=self.last_edge_removed[2]['weight'])
            self._update_graph(
                self._get_predecessor_edges(self.last_edge_removed))

    def _update_graph(self, edge=None):
        self.G.remove_edge(*edge[: 2])
        self.last_edge_removed = edge

    def _get_predecessor_edges(self, edge):
        if not self.predecessor_edges:
            node = edge[0]
            if node == "Source":
                self.stop = True
                return edge
            self.predecessor_edges = [
                e for e in self.G.edges(
                    self.G.nbunch_iter(
                        [node] + list(self.G.predecessors(node))),
                    data=True)
                if e[1] == node and e != edge]
            self.predecessor_edges.sort(
                key=lambda x: x[2]['weight'])
        next_edge = self.predecessor_edges[-1]
        self.predecessor_edges.pop(-1)
        return next_edge

    @staticmethod
    def _edge_extract(edge):
        return np.array(edge[2]['res_cost'])



from __future__ import absolute_import
from __future__ import print_function

import random
import logging
import numpy as np
from itertools import repeat
from operator import add, sub
from collections import OrderedDict
# Local module imports
from cspy.label import Label
from cspy.preprocessing import check_and_preprocess

log = logging.getLogger(__name__)


class BiDirectional:
    """
    Implementation of the bidirectional labeling algorithm with dynamic
    half-way point (`Tilk 2017`_).
    Depending on the range of values for U, L, we get
    four different algorithms. See self.name_algorithm and Notes.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats
        :math:`[H_F, M_1, M_2, ..., M_{n\_res}]` upper bounds for resource
        usage (including initial forward stopping point).
        We must have ``len(max_res)`` :math:`\geq 2`.

    min_res : list of floats
        :math:`[H_B, L_1, L_2, ..., L_{n\_res}]` lower bounds for resource
        usage (including initial backward stopping point).
        We must have ``len(min_res)`` :math:`=` ``len(max_res)`` :math:`\geq 2`

    direc_in : string, optional
        preferred search direction.
        Either 'both', 'forward', or, 'backward'. Default : 'both'.

    preprocess : bool, optional
        enables preprocessing routine.

    REF_forward REF_backward : functions, optional
        non-additive resource extension functions.

    Returns
    -------
    list
        nodes in shortest path obtained.

    Notes
    -----
    The input graph must have a ``n_res`` attribute which must be
    :math:`\geq 2`. The edges in the graph must all have a ``res_cost``
    attribute.

    According to the inputs, four different algorithms can be implemented,
    if you'd like to check which algorithm your running given your resource
    limits, run ``.name_algorithm(U, L)`` for a log with the classification.
    If ``direction`` is not given the absolute resource limits have to be
    given:

    U : float, optional
        Upper bound for monotone resource
        (for algorithm classification purposes see Notes).

    L : float, optional
        Lower bound for monotone resource
        (for algorithm classification purposes see Notes).

    According to these, we have,

    -  :math:`H_F = H_B > U` or ``direc_in`` = 'forward': Monodirectional forward labeling algorithm
    -  :math:`L < H_F = H_B < U`: Bidirectional labeling algorithm with static halfway point
    -  :math:`H_F = H_B < L` or ``direc_in`` = 'backward': Monodirectional backward labeling algorithm
    -  :math:`U = H_F > H_B = L`: Bidirectional labeling algorithm with dynamic halfway point.

    Example
    -------
    To run the algorithm, create a :class:`BiDirectional` instance and call
    ``run``.

    .. code-block:: python

        >>> import cspy
        >>> import networkx as nx
        >>> G = nx.DiGraph(directed=True, n_res=2)
        >>> G.add_edge('Source', 'A', res_cost=[1, 2], weight=0)
        >>> G.add_edge('A', 'B', res_cost=[1, 0.3], weight=0)
        >>> G.add_edge('A', 'C', res_cost=[1, 0.1], weight=0)
        >>> G.add_edge('B', 'C', res_cost=[1, 3], weight=-10)
        >>> G.add_edge('B', 'Sink', res_cost=[1, 2], weight=10)
        >>> G.add_edge('C', 'Sink', res_cost=[1, 10], weight=0)
        >>> max_res, min_res = [4, 20], [1, 0]
        >>> path = BiDirectional(G, max_res, min_res, direction='both').run()
        >>> print(path)
        ['Source', 'A', 'B', 'C', 'Sink']

    .. _Tilk 2017: https://www.sciencedirect.com/science/article/pii/S0377221717302035
    """

    def __init__(self, G, max_res, min_res, direction='both',
                 preprocess=True, REF_forward=add, REF_backward=sub):
        # Check inputs and preprocess G unless option disabled
        self.G = check_and_preprocess(
            preprocess, G, max_res, min_res, direction, __name__)
        self.direc_in = direction
        self.max_res, self.min_res = max_res, min_res
        # init current forward and backward labels
        self.Label = {
            'forward': Label(
                0, 'Source', np.zeros(G.graph['n_res']), ['Source']),
            'backward': Label(0, 'Sink', max_res, ['Sink'])}
        # init forward and backward unprocessed labels
        self.unprocessed = {'forward': {}, 'backward': {}}
        # init final path
        self.finalpath = {'forward': ["Source"], 'backward': ["Sink"]}
        # Set REFs for dominance relations and feasibility checks
        Label._REF_forward = REF_forward
        Label._REF_backward = REF_backward
        # Log algorithm type
        # self.name_algorithm(U, L)

    def run(self):
        while self.Label['forward'] or self.Label['backward']:
            direc = self.get_direction()
            if direc:
                self.algorithm(direc)
                self._check_dominance(direc)
            elif not direc or self._terminate(direc):
                break
        return self._join_paths()

    #############
    # DIRECTION #
    #############
    def get_direction(self):
        if self.direc_in == 'both':
            if self.Label['forward'] and not self.Label['backward']:
                return 'forward'
            elif not self.Label['forward'] and self.Label['backward']:
                return 'backward'
            elif self.Label['forward'] and self.Label['backward']:
                return random.choice(['forward', 'backward'])
            else:  # if both are empty
                return
        else:
            if not self.Label['forward'] and not self.Label['backward']:
                return
            elif not self.Label[self.direc_in]:
                return
            else:
                return self.direc_in

    #############
    # ALGORITHM #
    #############
    def algorithm(self, direc):
        if direc == 'forward':  # forward
            idx = 0  # index for head node
            self.min_res[0] = max(self.min_res[0], min(
                self.Label[direc].res[0], self.max_res[0]))
        else:
            idx = 1  # index for tail node
            self.max_res[0] = min(self.max_res[0], max(
                self.Label[direc].res[0], self.min_res[0]))
        # Select edges with the same head/tail node as the current label node.
        edges = [e for e in self.G.edges(data=True)
                 if e[idx] == self.Label[direc].node]
        # If Label not been seen before, initialise a dict
        if self.Label[direc] not in self.unprocessed[direc]:
            self.unprocessed[direc][self.Label[direc]] = {}
        # Propagate current label along all suitable edges in current direction
        list(map(self._propagate_label, edges, repeat(direc, len(edges))))
        # Extend label
        self._get_next_label(direc)

    def _propagate_label(self, edge, direc):
        # Label propagation #
        weight, res_cost = edge[2]['weight'], edge[2]['res_cost']
        node = edge[1] if direc == 'forward' else edge[0]
        new_label = self.Label[direc].get_new_label(
            direc, weight, node, res_cost)
        if new_label.feasibility_check(
                self.max_res, self.min_res, direc):
            self.unprocessed[direc][self.Label[direc]][
                new_label] = new_label.path

    def _get_next_label(self, direc):
        # Label Extension #
        keys_to_pop = []
        for key, val in self.unprocessed[direc].items():
            if val:
                # Update next forward label with one with least weight
                next_label = min(val.keys(), key=lambda x: x.weight)
                # Remove it from the unprocessed labels
                self.unprocessed[direc][key].pop(next_label)
                self.Label[direc] = next_label
                if not self.unprocessed[direc][key]:
                    keys_to_pop.append(key)
                break
            else:
                keys_to_pop.extend([self.Label[direc], key])
        else:  # if no break
            next_label = min(
                self.unprocessed[direc].keys(), key=lambda x: x.weight)
            # Remove it from the unprocessed labels
            keys_to_pop.append(next_label)
            if self.Label[direc] == next_label:
                self._save_final_path(direc)
                keys_to_pop.append(self.Label[direc])
                self.Label[direc] = None
            else:
                self.Label[direc] = next_label
        # Remove all processed labels from unprocessed dict
        for k in list(set(keys_to_pop)):
            self.unprocessed[direc].pop(k, None)

    def _save_final_path(self, direc):
        if self.Label[direc]:
            self.finalpath[direc] = self.Label[direc].path

    ###############
    # TERMINATION #
    ###############
    def _terminate(self, direc):
        if self.direc_in == "both":
            if (self.finalpath['forward'] and self.finalpath['backward'] and
                    (self.finalpath['forward'][-1] ==
                        self.finalpath['backward'][-1])):
                return True
        else:
            if (self.finalpath['forward'][-1] == "Sink" and
                    not self.unprocessed['forward']):
                return True
            elif (self.finalpath['backward'][-1] == "Source" and
                  not self.unprocessed['backward']):
                return True
            if not self.Label[self.direc_in] and not self.G.edges():
                return True

    #############
    # DOMINANCE #
    #############
    def _check_dominance(self, direc):
        """ For all labels, check if it is dominated, or itself dominates other
        labels. If this is found to be the case, the dominated label is
        removed """
        for sub_dict in ({k: v} for k, v in self.unprocessed[direc].items()):
            k = list(sub_dict.keys())[0]  # call dict_keys object as a list
            for label in [key for v in sub_dict.values() for key in v.keys()]:
                if label.node == self.Label[direc].node:
                    if self.Label[direc].dominates(label, direc):
                        self.unprocessed[direc][k].pop(
                            label, None)
                        self.unprocessed[direc].pop(label, None)
                    elif label.dominates(self.Label[direc], direc):
                        self.unprocessed[direc][k].pop(self.Label[direc], None)
                        self.unprocessed[direc].pop(self.Label[direc], None)

    #################
    # PATH CHECKING #
    #################
    def _join_paths(self):
        # check if paths are eligible to be joined
        if self.direc_in == 'both' and (
                self.finalpath['forward'] and self.finalpath['backward']):
                # reverse order for backward path
            self.finalpath['backward'].reverse()
            return self._check_paths()
        else:
            if self.direc_in == 'backward':
                self.finalpath[self.direc_in].reverse()
            return self.finalpath[self.direc_in]

    def _check_paths(self):
        if (self.finalpath['forward'][-1] == 'Sink' and
                self.finalpath['backward'][0] != 'Source'):
            # if only forward path
            return self.finalpath['forward']
        elif (self.finalpath['backward'][0] == 'Source' and
              self.finalpath['forward'][-1] != 'Sink'):
            # if only backward path
            return self.finalpath['backward']
        elif (self.finalpath['backward'][0] == 'Source' and
              self.finalpath['forward'][-1] == 'Sink'):
            # if both full paths
            return random.choice(
                [self.finalpath['forward'], self.finalpath['backward']])
        else:
            # if combination of the two is required
            return list(OrderedDict.fromkeys(
                self.finalpath['forward'] + self.finalpath['backward']))

    ###########################
    # Classify Algorithm Type #
    ###########################
    def name_algorithm(self, U=None, L=None):
        if (U and L) or self.direc_in != "both":
            HF, HB = self.max_res[0], self.min_res[0]
            if HF == HB > U or self.direc_in == 'forward':
                log.info('Monodirectional forward labeling algorithm')
            elif HF == HB < L or self.direc_in == 'backward':
                log.info('Monodirectional backward labeling algorithm')
            elif L < HF == HB < U:
                log.info('Bidirectional labeling algorithm with' +
                         ' static halfway point')
            elif U == HF > HB == L:
                log.info('Bidirectional labeling algorithm with' +
                         ' dynamic halfway point.')
